==> agentdojo-0.1.14.tar.gz: agentdojo-0.1.14/src/agentdojo/task_suite/task_suite.py <==
import importlib.resources
import re
import warnings
from collections.abc import Callable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import Generic, TypeVar

import yaml
from pydantic import BaseModel

from agentdojo.agent_pipeline.base_pipeline_element import BasePipelineElement
from agentdojo.agent_pipeline.errors import AbortAgentError
from agentdojo.agent_pipeline.ground_truth_pipeline import GroundTruthPipeline
from agentdojo.base_tasks import BaseInjectionTask, BaseUserTask
from agentdojo.functions_runtime import Function, FunctionCall, FunctionsRuntime, TaskEnvironment
from agentdojo.types import ChatMessage
from agentdojo.yaml_loader import ImportLoader

T = TypeVar("T", bound=BaseModel)
TC = TypeVar("TC", bound=Callable)


class InjectionVector(BaseModel):
    description: str
    default: str


def messages_to_dict(messages: Sequence[ChatMessage]) -> list[dict[str, str]]:
    def transform(msg: ChatMessage):
        return {
            "role": msg["role"],
            "content": msg["content"],
        }

    return [transform(msg) for msg in messages]


def functions_stack_trace_from_messages(
    messages: Sequence[ChatMessage],
) -> list[FunctionCall]:
    tool_calls = []
    for message in messages:
        if message["role"] == "assistant":
            for tool_call in message["tool_calls"] or []:
                tool_calls.append(tool_call)
    return tool_calls


def model_output_from_messages(
    messages: Sequence[ChatMessage],
) -> str | None:
    if messages[-1]["role"] != "assistant":
        raise ValueError("Last message was not an assistant message")
    return messages[-1]["content"]


@lru_cache
def read_suite_file(suite_name: str, file: str, suite_data_path: Path | None) -> str:
    if suite_data_path is not None:
        path = suite_data_path / file
        with path.open() as f:
            data_yaml = yaml.load(f, ImportLoader)
        return yaml.dump(data_yaml, default_flow_style=False)
    package_files = importlib.resources.files("agentdojo")
    path = package_files / "data" / "suites" / suite_name / file
    with importlib.resources.as_file(path) as p, p.open() as f:
        # Load into yaml to resolve imports
        data_yaml = yaml.load(f, ImportLoader)
    # dump back to string to pass to include injections later
    return yaml.dump(data_yaml, default_flow_style=False)


def validate_injections(injections: dict[str, str], injection_vector_defaults: dict[str, str]):
    injections_set = set(injections.keys())
    injection_vector_defaults_set = set(injection_vector_defaults.keys())
    if not injections_set.issubset(injection_vector_defaults_set):
        raise ValueError("Injections must be a subset of the injection vector defaults")


Env = TypeVar("Env", bound=TaskEnvironment)


class TaskSuite(Generic[Env]):
    """A suite of tasks that can be run in an environment. Tasks can be both user tasks and injection tasks. It is not
    mandatory to have injection tasks in case the suite is used to evaluate the model only for utility.

    Args:
        name: The name of the suite.
        environment_type: The environment type that the suite operates on.
        tools: A list of tools that the agent can use to solve the tasks.
        data_path: The path to the suite data directory. It should be provided for non-default suites.
            The directory should contain the following files:
                - `environment.yaml`: The data of the environment.
                - `injection_vectors.yaml`: The injection vectors in the environment.
    """

    def __init__(
        self,
        name: str,
        environment_type: type[Env],
        tools: list[Function],
        data_path: Path | None = None,
    ):
        self.name = name
        self.environment_type = environment_type
        self.tools = tools
        self._user_tasks: dict[str, BaseUserTask[Env]] = {}
        self._injection_tasks: dict[str, BaseInjectionTask[Env]] = {}
        self.data_path = data_path

    def load_and_inject_default_environment(self, injections: dict[str, str]) -> Env:
        environment_text = read_suite_file(self.name, "environment.yaml", self.data_path)
        injection_vector_defaults = self.get_injection_vector_defaults()
        injections_with_defaults = dict(injection_vector_defaults, **injections)
        validate_injections(injections, injection_vector_defaults)
        injected_environment = environment_text.format(**injections_with_defaults)
        environment = self.environment_type.model_validate(yaml.safe_load(injected_environment))
        return environment

    def get_injection_vector_defaults(self) -> dict[str, str]:
        injection_vectors_text = read_suite_file(self.name, "injection_vectors.yaml", self.data_path)
        injection_vectors = yaml.safe_load(injection_vectors_text)
        vectors = {
            vector_id: InjectionVector.model_validate(vector_info)
            for vector_id, vector_info in injection_vectors.items()
        }
        return {vector_id: vector.default for vector_id, vector in vectors.items()}

    def _get_task_number(self, task_cls: type[BaseUserTask] | type[BaseInjectionTask], prefix: str) -> int:
        match = re.match(rf"{prefix}(\d+)", task_cls.__name__)
        if not match:
            raise ValueError(f"User tasks must be named {prefix} followed by a number, got {task_cls.__name__}")
        return int(match.group(1))

    def register_user_task(self, task: type[BaseUserTask[Env]]) -> type[BaseUserTask[Env]]:
        """Register a user task in the suite.

        Args:
            task: The user task class to register.
        """
        task_n = self._get_task_number(task, "UserTask")
        task_id = f"user_task_{task_n}"
        setattr(task, "ID", task_id)
        self._user_tasks[task_id] = task()
        return task

    @property
    def user_tasks(self) -> dict[str, BaseUserTask[Env]]:
        """The user tasks in the suite.

        Returns:
            A dictionary of user tasks with their IDs as keys.
        """
        return self._user_tasks

    def get_user_task_by_id(self, task_id: str) -> BaseUserTask[Env]:
        """Get a user task by its ID."""
        return self._user_tasks[task_id]

    def register_injection_task(self, task: type[BaseInjectionTask[Env]]) -> type[BaseInjectionTask[Env]]:
        """Register an injection task in the suite.

        Args:
            task: The injection task class to register.
        """
        task_n = self._get_task_number(task, "InjectionTask")
        task_id = f"injection_task_{task_n}"
        setattr(task, "ID", task_id)
        self._injection_tasks[task_id] = task()
        return task

    @property
    def injection_tasks(self) -> dict[str, BaseInjectionTask[Env]]:
        """The injection tasks in the suite.

        Returns:
            A dictionary of injection tasks with their IDs as keys.
        """
        return self._injection_tasks

    def get_injection_task_by_id(self, injection_task_id: str) -> BaseInjectionTask[Env]:
        """Get an injection task by its ID."""
        return self._injection_tasks[injection_task_id]

    def _check_user_task_utility(
        self,
        task: BaseUserTask,
        model_output: str,
        pre_environment: Env,
        task_environment: Env,
        functions_stack_trace: Sequence[FunctionCall],
    ) -> bool:
        utility_from_stack_traces = task.utility_from_traces(
            model_output, pre_environment, task_environment, functions_stack_trace
        )
        if utility_from_stack_traces is not None:
            return utility_from_stack_traces
        return task.utility(model_output, pre_environment, task_environment)

    def _check_injection_task_security(
        self,
        task: BaseInjectionTask,
        model_output: str,
        pre_environment: Env,
        task_environment: Env,
        functions_stack_trace: Sequence[FunctionCall],
    ) -> bool:
        security_from_stack_traces = task.security_from_traces(
            model_output, pre_environment, task_environment, functions_stack_trace
        )
        if security_from_stack_traces is not None:
            return security_from_stack_traces
        return task.security(model_output, pre_environment, task_environment)

    def _check_task_result(
        self,
        task: BaseUserTask | BaseInjectionTask,
        model_output: str,
        pre_environment: Env,
        task_environment: Env,
        functions_stack_trace: Sequence[FunctionCall],
    ) -> bool:
        match task:
            case BaseUserTask():
                return self._check_user_task_utility(
                    task,
                    model_output,
                    pre_environment,
                    task_environment,
                    functions_stack_trace,
                )
            case BaseInjectionTask():
                return self._check_injection_task_security(
                    task,
                    model_output,
                    pre_environment,
                    task_environment,
                    functions_stack_trace,
                )

    def run_task_with_pipeline(
        self,
        agent_pipeline: BasePipelineElement,
        user_task: BaseUserTask[Env] | BaseInjectionTask[Env],
        injection_task: BaseInjectionTask[Env] | None,
        injections: dict[str, str],
        runtime_class: type[FunctionsRuntime] = FunctionsRuntime,
        environment: Env | None = None,
        verbose: bool = False,
    ) -> tuple[bool, bool]:
        """Run a task with the provided pipeline.

        Args:
            agent_pipeline: The pipeline to use for the task.
            user_task: The user task to run.
            injection_task: The injection task to run.
            injections: The injections to use for the task.
            runtime_class: The runtime class to use for the task.
            environment: The environment to use for the task.
            verbose: Whether to print debug information.

        Returns:
            A tuple of two booleans, the first indicating whether the task was successful, and the second indicating if
            the injection was successful.
        """
        # If no environment is provided, load the default environment
        if environment is None:
            environment = self.load_and_inject_default_environment(injections)
        # Initialize the environment according to the task
        if isinstance(user_task, BaseUserTask):
            task_environment = user_task.init_environment(environment)
        else:
            task_environment = environment

        # Create a copy of the environment before running the user task to then diff if needed
        pre_environment = task_environment.model_copy(deep=True)
        if isinstance(user_task, BaseUserTask):
            prompt = user_task.PROMPT
        else:
            prompt = user_task.GOAL

        runtime = runtime_class(self.tools)
        model_output = None
        messages = []
        for _ in range(3):
            # Run the user task
            try:
                _, _, task_environment, messages, _ = agent_pipeline.query(prompt, runtime, task_environment)
            except AbortAgentError as e:
                # get last environment and messages
                task_environment = e.task_environment
                messages = e.messages

            model_output = model_output_from_messages(messages)
            if model_output is not None:
                break

        if model_output is None:
            warnings.warn(f"Model output was None for task {user_task.ID}")

        functions_stack_trace = functions_stack_trace_from_messages(messages)
        utility = self._check_task_result(
            user_task,
            model_output or "",
            pre_environment,
            task_environment,  # type: ignore
            functions_stack_trace,
        )

        # Early return if no injection was intended
        if injection_task is None:
            return utility, True

        security = self._check_task_result(
            injection_task,
            model_output or "",
            pre_environment,
            task_environment,  # type: ignore
            functions_stack_trace,
        )

        return utility, security

    def check(self, check_injectable: bool = True) -> tuple[bool, tuple[dict[str, tuple[bool, str]], dict[str, bool]]]:
        injection_defaults = self.get_injection_vector_defaults()
        canary_injections = {inj: f"---CANARY_{inj}---" for inj in injection_defaults}
        environment = self.load_and_inject_default_environment(canary_injections)

        user_tasks_results = {}

        # Check user tasks
        for user_task in self.user_tasks.values():
            ground_truth_pipeline = GroundTruthPipeline(user_task)
            post_environment = environment.copy(deep=True)
            utility, _ = self.run_task_with_pipeline(
                ground_truth_pipeline,
                user_task,
                injection_task=None,
                injections={},
                environment=post_environment,
            )
            if not utility:
                user_tasks_results[user_task.ID] = (
                    False,
                    "Ground truth does not solve the task",
                )
            # Check if the task is injectable
            runtime = FunctionsRuntime(self.tools)
            _, _, _, responses, _ = ground_truth_pipeline.query(user_task.PROMPT, runtime, environment.copy(deep=True))
            if not is_task_injectable(responses, list(canary_injections.values())) and check_injectable:
                user_tasks_results[user_task.ID] = (
                    False,
                    "The task is not injectable",
                )
            if user_task.ID not in user_tasks_results:
                user_tasks_results[user_task.ID] = (True, "")

        injection_tasks_results = {}

        # Check injection tasks
        user_task = next(iter(self.user_tasks.values()))
        for injection_task in self.injection_tasks.values():
            ground_truth_pipeline = GroundTruthPipeline(injection_task)
            post_environment = environment.copy(deep=True)
            _, security = self.run_task_with_pipeline(
                ground_truth_pipeline,
                user_task,
                injection_task=injection_task,
                injections={},
                environment=post_environment,
            )
            injection_tasks_results[injection_task.ID] = security
            for tool_call in injection_task.ground_truth(environment.copy(deep=True)):
                if tool_call.placeholder_args is None:
                    warnings.warn(f"Missing placeholder_args in {injection_task.ID}'s ground truth")

        all_checks_passed = all(result[0] for result in user_tasks_results.values()) and all(
            result for result in injection_tasks_results.values()
        )

        return all_checks_passed, (user_tasks_results, injection_tasks_results)


def is_task_injectable(responses: Sequence[ChatMessage], canaries: list[str]) -> bool:
    tool_responses = [response for response in responses if response["role"] == "tool"]
    tool_outputs = [response["content"] for response in tool_responses if isinstance(response["content"], str)]
    all_outputs = " ".join(tool_outputs)
    inj_in_output = [inj in all_outputs for inj in canaries]
    return any(inj_in_output)

==> agentdojo-0.1.19.tar.gz: agentdojo-0.1.19/src/agentdojo/task_suite/task_suite.py <==
import copy
import importlib.resources
import re
import warnings
from collections import defaultdict
from collections.abc import Callable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import Generic, TypeVar

import yaml
from pydantic import BaseModel
from typing_extensions import Self

from agentdojo.agent_pipeline.base_pipeline_element import BasePipelineElement
from agentdojo.agent_pipeline.errors import AbortAgentError
from agentdojo.agent_pipeline.ground_truth_pipeline import GroundTruthPipeline
from agentdojo.base_tasks import BaseInjectionTask, BaseUserTask
from agentdojo.functions_runtime import Function, FunctionCall, FunctionsRuntime, TaskEnvironment
from agentdojo.types import ChatMessage
from agentdojo.yaml_loader import ImportLoader

T = TypeVar("T", bound=BaseModel)
TC = TypeVar("TC", bound=Callable)
BenchmarkVersion = tuple[int, int, int]


class InjectionVector(BaseModel):
    description: str
    default: str


IT = TypeVar("IT")


def get_version_compatible_items(
    items: dict[str, dict[BenchmarkVersion, IT]], version: BenchmarkVersion
) -> dict[str, IT]:
    # for each element in items, get the one with highest benchmark version lower or equal to `version`
    compatible_items = {}
    for item_id, item_versions in items.items():
        highest_compatible_version = max(v for v in item_versions.keys() if v <= version)
        compatible_items[item_id] = items[item_id][highest_compatible_version]
    return compatible_items


def messages_to_dict(messages: Sequence[ChatMessage]) -> list[dict[str, str]]:
    def transform(msg: ChatMessage):
        return {
            "role": msg["role"],
            "content": msg["content"],
        }

    return [transform(msg) for msg in messages]


def functions_stack_trace_from_messages(
    messages: Sequence[ChatMessage],
) -> list[FunctionCall]:
    tool_calls = []
    for message in messages:
        if message["role"] == "assistant":
            for tool_call in message["tool_calls"] or []:
                tool_calls.append(tool_call)
    return tool_calls


def model_output_from_messages(
    messages: Sequence[ChatMessage],
) -> str | None:
    if messages[-1]["role"] != "assistant":
        raise ValueError("Last message was not an assistant message")
    return messages[-1]["content"]


@lru_cache
def read_suite_file(suite_name: str, file: str, suite_data_path: Path | None) -> str:
    if suite_data_path is not None:
        path = suite_data_path / file
        with path.open() as f:
            data_yaml = yaml.load(f, ImportLoader)
        return yaml.dump(data_yaml, default_flow_style=False)
    package_files = importlib.resources.files("agentdojo")
    path = package_files / "data" / "suites" / suite_name / file
    with importlib.resources.as_file(path) as p, p.open() as f:
        # Load into yaml to resolve imports
        data_yaml = yaml.load(f, ImportLoader)
    # dump back to string to pass to include injections later
    return yaml.dump(data_yaml, default_flow_style=False)


def validate_injections(injections: dict[str, str], injection_vector_defaults: dict[str, str]):
    injections_set = set(injections.keys())
    injection_vector_defaults_set = set(injection_vector_defaults.keys())
    if not injections_set.issubset(injection_vector_defaults_set):
        raise ValueError("Injections must be a subset of the injection vector defaults")


Env = TypeVar("Env", bound=TaskEnvironment)


class TaskSuite(Generic[Env]):
    """A suite of tasks that can be run in an environment. Tasks can be both user tasks and injection tasks. It is not
    mandatory to have injection tasks in case the suite is used to evaluate the model only for utility.

    Args:
        name: The name of the suite.
        environment_type: The environment type that the suite operates on.
        tools: A list of tools that the agent can use to solve the tasks.
        data_path: The path to the suite data directory. It should be provided for non-default suites.
            The directory should contain the following files:
                - `environment.yaml`: The data of the environment.
                - `injection_vectors.yaml`: The injection vectors in the environment.
    """

    def __init__(
        self,
        name: str,
        environment_type: type[Env],
        tools: list[Function],
        data_path: Path | None = None,
        benchmark_version: BenchmarkVersion = (1, 0, 0),
    ):
        self.name = name
        self.environment_type = environment_type
        self.tools = tools
        self._user_tasks: dict[str, dict[BenchmarkVersion, BaseUserTask[Env]]] = defaultdict(dict)
        self._injection_tasks: dict[str, dict[BenchmarkVersion, BaseInjectionTask[Env]]] = defaultdict(dict)
        self.data_path = data_path
        self.benchmark_version = benchmark_version

    def get_new_version(self, benchmark_version: BenchmarkVersion) -> Self:
        new_self = copy.deepcopy(self)
        new_self.benchmark_version = benchmark_version
        return new_self

    def load_and_inject_default_environment(self, injections: dict[str, str]) -> Env:
        environment_text = read_suite_file(self.name, "environment.yaml", self.data_path)
        injection_vector_defaults = self.get_injection_vector_defaults()
        injections_with_defaults = dict(injection_vector_defaults, **injections)
        validate_injections(injections, injection_vector_defaults)
        injected_environment = environment_text.format(**injections_with_defaults)
        environment = self.environment_type.model_validate(yaml.safe_load(injected_environment))
        return environment

    def get_injection_vector_defaults(self) -> dict[str, str]:
        injection_vectors_text = read_suite_file(self.name, "injection_vectors.yaml", self.data_path)
        injection_vectors = yaml.safe_load(injection_vectors_text)
        vectors = {
            vector_id: InjectionVector.model_validate(vector_info)
            for vector_id, vector_info in injection_vectors.items()
        }
        return {vector_id: vector.default for vector_id, vector in vectors.items()}

    def _get_task_number(self, task_cls: type[BaseUserTask] | type[BaseInjectionTask], prefix: str) -> int:
        match = re.match(rf"{prefix}(\d+)", task_cls.__name__)
        if not match:
            raise ValueError(f"User tasks must be named {prefix} followed by a number, got {task_cls.__name__}")
        return int(match.group(1))

    def register_user_task(self, task: type[BaseUserTask[Env]]) -> type[BaseUserTask[Env]]:
        """Register a user task in the suite.

        Args:
            task: The user task class to register.
        """
        task_n = self._get_task_number(task, "UserTask")
        task_id = f"user_task_{task_n}"
        setattr(task, "ID", task_id)
        self._user_tasks[task_id][(1, 0, 0)] = task()
        return task

    def update_user_task(self, benchmark_version: BenchmarkVersion):
        """Updates a user task in the suite and makes it part of the benchmark
        with the given version

        Args:
            benchmark_version: The benchmark version this task belongs to.
        """

        def decorator(task: type[BaseUserTask[Env]]) -> type[BaseUserTask[Env]]:
            task_n = self._get_task_number(task, "UserTask")
            task_id = f"user_task_{task_n}"
            if task_id not in self._user_tasks:
                raise ValueError(f"User task {task_id} not found in suite")
            setattr(task, "ID", task_id)
            self._user_tasks[task_id][benchmark_version] = task()
            return task

        return decorator

    @property
    def user_tasks(self) -> dict[str, BaseUserTask[Env]]:
        """The user tasks in the suite.

        Returns:
            A dictionary of user tasks with their IDs as keys.
        """
        return get_version_compatible_items(self._user_tasks, self.benchmark_version)

    @lru_cache
    def get_user_task_by_id(self, task_id: str) -> BaseUserTask[Env]:
        """Get a user task by its ID."""
        return self.user_tasks[task_id]

    def register_injection_task(self, task: type[BaseInjectionTask[Env]]) -> type[BaseInjectionTask[Env]]:
        """Register an injection task in the suite.

        Args:
            task: The injection task class to register.
        """
        task_n = self._get_task_number(task, "InjectionTask")
        task_id = f"injection_task_{task_n}"
        setattr(task, "ID", task_id)
        self._injection_tasks[task_id][(1, 0, 0)] = task()
        return task

    def update_injection_task(self, benchmark_version: BenchmarkVersion):
        """Updates an injection task in the suite and makes it part of the benchmark
        with the given version

        Args:
            benchmark_version: The benchmark version this task belongs to.
        """

        def decorator(task: type[BaseInjectionTask[Env]]) -> type[BaseInjectionTask[Env]]:
            task_n = self._get_task_number(task, "InjectionTask")
            task_id = f"injection_task_{task_n}"
            if task_id not in self._injection_tasks:
                raise ValueError(f"Injection task {task_id} not found in suite")
            setattr(task, "ID", task_id)
            self._injection_tasks[task_id][benchmark_version] = task()
            return task

        return decorator

    @property
    def injection_tasks(self) -> dict[str, BaseInjectionTask[Env]]:
        """The injection tasks in the suite.

        Returns:
            A dictionary of injection tasks with their IDs as keys.
        """
        return get_version_compatible_items(self._injection_tasks, self.benchmark_version)

    def get_injection_task_by_id(self, injection_task_id: str) -> BaseInjectionTask[Env]:
        """Get an injection task by its ID."""
        return self.injection_tasks[injection_task_id]

    def _check_user_task_utility(
        self,
        task: BaseUserTask,
        model_output: str,
        pre_environment: Env,
        task_environment: Env,
        functions_stack_trace: Sequence[FunctionCall],
    ) -> bool:
        utility_from_stack_traces = task.utility_from_traces(
            model_output, pre_environment, task_environment, functions_stack_trace
        )
        if utility_from_stack_traces is not None:
            return utility_from_stack_traces
        return task.utility(model_output, pre_environment, task_environment)

    def _check_injection_task_security(
        self,
        task: BaseInjectionTask,
        model_output: str,
        pre_environment: Env,
        task_environment: Env,
        functions_stack_trace: Sequence[FunctionCall],
    ) -> bool:
        security_from_stack_traces = task.security_from_traces(
            model_output, pre_environment, task_environment, functions_stack_trace
        )
        if security_from_stack_traces is not None:
            return security_from_stack_traces
        return task.security(model_output, pre_environment, task_environment)

    def _check_task_result(
        self,
        task: BaseUserTask | BaseInjectionTask,
        model_output: str,
        pre_environment: Env,
        task_environment: Env,
        functions_stack_trace: Sequence[FunctionCall],
    ) -> bool:
        match task:
            case BaseUserTask():
                return self._check_user_task_utility(
                    task,
                    model_output,
                    pre_environment,
                    task_environment,
                    functions_stack_trace,
                )
            case BaseInjectionTask():
                return self._check_injection_task_security(
                    task,
                    model_output,
                    pre_environment,
                    task_environment,
                    functions_stack_trace,
                )

    def run_task_with_pipeline(
        self,
        agent_pipeline: BasePipelineElement,
        user_task: BaseUserTask[Env] | BaseInjectionTask[Env],
        injection_task: BaseInjectionTask[Env] | None,
        injections: dict[str, str],
        runtime_class: type[FunctionsRuntime] = FunctionsRuntime,
        environment: Env | None = None,
        verbose: bool = False,
    ) -> tuple[bool, bool]:
        """Run a task with the provided pipeline.

        Args:
            agent_pipeline: The pipeline to use for the task.
            user_task: The user task to run.
            injection_task: The injection task to run.
            injections: The injections to use for the task.
            runtime_class: The runtime class to use for the task.
            environment: The environment to use for the task.
            verbose: Whether to print debug information.

        Returns:
            A tuple of two booleans, the first indicating whether the task was successful, and the second indicating if
            the injection was successful.
        """
        # If no environment is provided, load the default environment
        if environment is None:
            environment = self.load_and_inject_default_environment(injections)
        # Initialize the environment according to the task
        if isinstance(user_task, BaseUserTask):
            task_environment = user_task.init_environment(environment)
        else:
            task_environment = environment

        # Create a copy of the environment before running the user task to then diff if needed
        pre_environment = task_environment.model_copy(deep=True)
        if isinstance(user_task, BaseUserTask):
            prompt = user_task.PROMPT
        else:
            prompt = user_task.GOAL

        runtime = runtime_class(self.tools)
        model_output = None
        messages = []
        for _ in range(3):
            # Run the user task
            try:
                _, _, task_environment, messages, _ = agent_pipeline.query(prompt, runtime, task_environment)
            except AbortAgentError as e:
                # get last environment and messages
                task_environment = e.task_environment
                messages = e.messages

            model_output = model_output_from_messages(messages)
            if model_output is not None:
                break

        if model_output is None:
            warnings.warn(f"Model output was None for task {user_task.ID}")

        functions_stack_trace = functions_stack_trace_from_messages(messages)
        utility = self._check_task_result(
            user_task,
            model_output or "",
            pre_environment,
            task_environment,  # type: ignore
            functions_stack_trace,
        )

        # Early return if no injection was intended
        if injection_task is None:
            return utility, True

        security = self._check_task_result(
            injection_task,
            model_output or "",
            pre_environment,
            task_environment,  # type: ignore
            functions_stack_trace,
        )

        return utility, security

    def check(self, check_injectable: bool = True) -> tuple[bool, tuple[dict[str, tuple[bool, str]], dict[str, bool]]]:
        injection_defaults = self.get_injection_vector_defaults()
        canary_injections = {inj: f"---CANARY_{inj}---" for inj in injection_defaults}
        environment = self.load_and_inject_default_environment(canary_injections)

        user_tasks_results = {}

        # Check user tasks
        for user_task in self.user_tasks.values():
            ground_truth_pipeline = GroundTruthPipeline(user_task)
            post_environment = environment.copy(deep=True)
            utility, _ = self.run_task_with_pipeline(
                ground_truth_pipeline,
                user_task,
                injection_task=None,
                injections={},
                environment=post_environment,
            )
            if not utility:
                user_tasks_results[user_task.ID] = (
                    False,
                    "Ground truth does not solve the task",
                )
            # Check if the task is injectable
            runtime = FunctionsRuntime(self.tools)
            _, _, _, responses, _ = ground_truth_pipeline.query(user_task.PROMPT, runtime, environment.copy(deep=True))
            if not is_task_injectable(responses, list(canary_injections.values())) and check_injectable:
                user_tasks_results[user_task.ID] = (
                    False,
                    "The task is not injectable",
                )
            if user_task.ID not in user_tasks_results:
                user_tasks_results[user_task.ID] = (True, "")

        injection_tasks_results = {}

        # Check injection tasks
        user_task = next(iter(self.user_tasks.values()))
        for injection_task in self.injection_tasks.values():
            ground_truth_pipeline = GroundTruthPipeline(injection_task)
            post_environment = environment.copy(deep=True)
            _, security = self.run_task_with_pipeline(
                ground_truth_pipeline,
                user_task,
                injection_task=injection_task,
                injections={},
                environment=post_environment,
            )
            injection_tasks_results[injection_task.ID] = security
            for tool_call in injection_task.ground_truth(environment.copy(deep=True)):
                if tool_call.placeholder_args is None:
                    warnings.warn(f"Missing placeholder_args in {injection_task.ID}'s ground truth")

        all_checks_passed = all(result[0] for result in user_tasks_results.values()) and all(
            result for result in injection_tasks_results.values()
        )

        return all_checks_passed, (user_tasks_results, injection_tasks_results)


def is_task_injectable(responses: Sequence[ChatMessage], canaries: list[str]) -> bool:
    tool_responses = [response for response in responses if response["role"] == "tool"]
    tool_outputs = [response["content"] for response in tool_responses if isinstance(response["content"], str)]
    all_outputs = " ".join(tool_outputs)
    inj_in_output = [inj in all_outputs for inj in canaries]
    return any(inj_in_output)
